// src/scheduler.ts
function createScheduler(tick) {
  const queue = /* @__PURE__ */ new Set();
  let flushing = false;
  const flush_ = () => {
    for (const task of Array.from(queue).slice()) {
      task();
      queue.delete(task);
    }
  };
  return {
    tick,
    b: flush_,
    a: (task) => {
      queue.add(task);
      if (!flushing) {
        flushing = true;
        tick(() => {
          flushing = false;
          flush_();
        });
      }
    }
  };
}
function createFrameTicker() {
  return (cb) => requestAnimationFrame(cb);
}
function createMicroTicker() {
  const promise = Promise.resolve();
  return (cb) => void promise.then(cb);
}

// src/model.ts
var createLink = (target, key) => [target, key];
var GET = 0;
var MODIFY = 1;
function get(map, key, _default) {
  let value;
  map.has(key) ? value = map.get(key) : map.set(key, value = _default());
  return value;
}
function createEvents() {
  const onAll = [
    /* @__PURE__ */ new Set(),
    /* @__PURE__ */ new Set()
  ];
  const onLink = /* @__PURE__ */ new Map();
  const copy = /* @__PURE__ */ new Set();
  const getOnLink = (type, link) => get(get(onLink, link[0], () => /* @__PURE__ */ new Map()), link[1], () => [/* @__PURE__ */ new Set(), /* @__PURE__ */ new Set()])[type];
  return {
    emit(type, link) {
      for (const events of copy)
        events.emit(...arguments);
      for (const listener of onAll[type])
        listener(link);
      for (const listener of getOnLink(type, link))
        listener();
    },
    pipe(to) {
      copy.add(to);
    },
    on(type, link, listener) {
      const target = link ? getOnLink(type, link) : onAll[type];
      target.add(listener);
      return () => target.delete(listener);
    }
  };
}
var ATTACH = /* @__PURE__ */ new WeakMap();
var EXTENDERS = /* @__PURE__ */ new WeakMap();
function attach(target, property) {
  let set;
  if (set = ATTACH.get(target))
    set.add(property);
  else
    ATTACH.set(target, /* @__PURE__ */ new Set([property]));
  ignore(target, property);
}
function extend(extender) {
  return (target, property) => {
    let map;
    if (map = EXTENDERS.get(target)) {
      if (map.has(property))
        throw new Error("can only extend once");
      else
        map.set(property, extender);
    } else {
      EXTENDERS.set(target, /* @__PURE__ */ new Map([[property, extender]]));
    }
  };
}
function ignore(...args) {
  extend(() => {
  })(...args);
}
var Model = class {
  $observe_(link, listener) {
    return this.$e.on(MODIFY, link, listener);
  }
  $schedule(task, dedup) {
    if (dedup) {
      let set;
      if (this.$q.size === 0)
        this.$s.a(() => this.$q.clear());
      if (set = this.$q.get(dedup[0])) {
        if (set.has(dedup[1]))
          return;
        else
          set.add(dedup[1]);
      } else {
        this.$q.set(dedup[0], /* @__PURE__ */ new Set([dedup[1]]));
      }
    }
    this.$s.a(task);
  }
  $scheduleEmit_(type, link) {
    this.$schedule(() => this.$e.emit(type, link), link);
  }
  $tick() {
    this.$s.b();
  }
  $magic(callback) {
    const links = [];
    const dispose = this.$e.on(GET, null, (link) => links.push(link));
    callback();
    dispose();
    return (listener) => {
      const observers = links.map((link) => this.$observe_(link, listener));
      return () => {
        for (const revoke of observers)
          revoke();
      };
    };
  }
  $observe(target, listener) {
    return {
      revoke: this.$observe_(target, listener)
    };
  }
  $notify(target) {
    throw new Error("proxy was bypassed");
  }
  $ref(target) {
    throw new Error("proxy was bypassed");
  }
  $silent(target) {
    throw new Error("proxy was bypassed");
  }
  constructor(options) {
    var _a;
    this.$e = createEvents();
    this.$q = /* @__PURE__ */ new Map();
    this.$ticker = (_a = options == null ? void 0 : options.ticker) != null ? _a : createMicroTicker();
    this.$s = createScheduler(this.$ticker);
    const attach2 = ATTACH.get(Object.getPrototypeOf(this));
    const extenders = EXTENDERS.get(Object.getPrototypeOf(this));
    const extend2 = (object, isRoot) => {
      let returnLink = false;
      let silent = false;
      const _compted = /* @__PURE__ */ new WeakMap();
      const asComputed = (target, key, receiver, descriptior) => {
        if (_compted.has(descriptior.get))
          return _compted.get(descriptior.get)();
        let value;
        let revoke;
        let dirty = true;
        const get2 = () => {
          if (dirty) {
            revoke == null ? void 0 : revoke();
            dirty = false;
            let next;
            const observe = this.$magic(() => next = Reflect.get(target, key, receiver));
            if (next !== value)
              value = next;
            revoke = observe(() => {
              dirty = true;
              this.$scheduleEmit_(MODIFY, createLink(target, key));
            });
          }
          return value;
        };
        _compted.set(descriptior.get, get2);
        return get2();
      };
      return new Proxy(object, {
        get: (target, key, receiver) => {
          if (isRoot) {
            switch (key) {
              case "$observe": {
                returnLink = true;
                break;
              }
              case "$ref": {
                returnLink = true;
                return (link) => ({
                  get: () => Reflect.get(...link),
                  set: (value2) => void Reflect.set(...link, value2),
                  observe: (listener) => ({
                    revoke: this.$observe_(link, listener)
                  })
                });
              }
              case "$notify": {
                returnLink = true;
                return (link) => this.$scheduleEmit_(MODIFY, link);
              }
              case "$silent": {
                silent = true;
                return (value2) => {
                  silent = false;
                  return value2;
                };
              }
            }
            if (typeof key === "string" && key.startsWith("$"))
              return Reflect.get(target, key, receiver);
          }
          if (returnLink) {
            returnLink = false;
            return createLink(target, key);
          }
          const descriptior = Object.getOwnPropertyDescriptor(Object.getPrototypeOf(target), key);
          if (descriptior == null ? void 0 : descriptior.get)
            return asComputed(target, key, receiver, descriptior);
          const value = Reflect.get(target, key, receiver);
          if (!silent) {
            let extender;
            if (extender = extenders == null ? void 0 : extenders.get(key)) {
              extender(() => this.$e.emit(GET, createLink(target, key)));
            } else {
              this.$e.emit(GET, createLink(target, key));
            }
          }
          if (typeof value === "object" && value !== null)
            return extend2(value, false);
          return value;
        },
        set: (target, key, value, receiver) => {
          const succeeded = Reflect.set(target, key, value, receiver);
          if (isRoot && (attach2 == null ? void 0 : attach2.has(key))) {
            const model = value;
            model.$e.pipe(this.$e);
          }
          if (!silent) {
            let extender;
            if (extender = extenders == null ? void 0 : extenders.get(key)) {
              extender(() => this.$scheduleEmit_(MODIFY, createLink(target, key)));
            } else {
              this.$scheduleEmit_(MODIFY, createLink(target, key));
            }
          }
          return succeeded;
        },
        defineProperty: (target, property, attributes) => {
          var _a2, _b;
          if (isRoot && (attach2 == null ? void 0 : attach2.has(property))) {
            const model = (_b = attributes.value) != null ? _b : (_a2 = attributes.get) == null ? void 0 : _a2.call(attributes);
            if (model)
              model.$e.pipe(this.$e);
          }
          return Reflect.defineProperty(target, property, attributes);
        }
      });
    };
    return extend2(this, true);
  }
};

// src/viewmodel.ts
var ViewModel = class extends Model {
  constructor(options = {}) {
    var _a;
    (_a = options.ticker) != null ? _a : options.ticker = createFrameTicker();
    super(options);
  }
  $computed(get2) {
    const listeners = /* @__PURE__ */ new Set();
    let value;
    let revoke;
    let dirty = true;
    return {
      get: () => {
        if (dirty) {
          revoke == null ? void 0 : revoke();
          dirty = false;
          let next;
          const observe = this.$magic(() => next = get2());
          if (next !== value)
            value = next;
          revoke = observe(() => {
            dirty = true;
            this.$schedule(() => {
              for (const listener of listeners)
                listener();
            });
          });
        }
        return value;
      },
      observe: (listener) => {
        listeners.add(listener);
        return {
          revoke() {
            listeners.delete(listener);
          }
        };
      },
      dispose: () => {
        revoke == null ? void 0 : revoke();
      }
    };
  }
};
export {
  Model,
  ViewModel,
  attach,
  extend,
  ignore
};
